@@ -1,8 +1,8 @@
 import torch
-from torch.utils.data import DataLoader, TensorDataset
-import torchsummary
+import numpy as np
+from torch.utils.data import DataLoader, TensorDataset, WeightedRandomSampler, Subset
 from collections import OrderedDict
-from torch.utils.data import random_split
+from .._summary import summary
 
 from .._training import Trainer
 from ..layers import KerasLayer
@@ -10,7 +10,9 @@ from ..metrics import Metric
 from ..metrics import _create_metric
 from ..losses import _create_loss
 from ..optimizers import _create_optimizer
-from ..utils import to_tensor
+from ..activations import _create_activation
+from ..utils import to_tensor, _get_num_workers
+
 
 class Model(torch.nn.Module):
     """
@@ -26,41 +28,56 @@ class Model(torch.nn.Module):
         self.model = model
         self.compiled = False
         self.built = False
+        self.input_shape = None
 
-        self.has_keras_layer = False
-        def check_keras_layer(m):
-            if isinstance(m, KerasLayer):
-                self.has_keras_layer = True
-        self.model.apply(check_keras_layer)
+        def dfs(m):
+            for child in m.children():
+                if isinstance(child, KerasLayer) or dfs(child):
+                    return True
+            return False
+        self._has_keras_layer = dfs(self)
 
     def forward(self, x):
-        return self.model.forward(x)
+        return self.model(x)
 
     def count_params(self) -> int:
         """Count the total number of scalars composing the weights."""
         return sum([p.numel() for p in self.parameters()])
 
+    def trainable_params(self):
+        """Return all trainable parameters of the model."""
+        return filter(lambda p: p.requires_grad, self.parameters())
 
     ########## keras-style methods below ##########
 
     @torch.no_grad()
     def build(self, input_shape, dtype=torch.float32):
         """Build the model when it contains `KerasLayer`."""
-        if self.has_keras_layer:
-            input_shape = [2] + list(input_shape)
-            probe_input = torch.zeros(size=input_shape).to(dtype=dtype)
-            self.model.forward(probe_input)
+        if self._has_keras_layer:
+            batch_shape = [2] + list(input_shape)
+            probe_input = torch.zeros(size=batch_shape).to(dtype=dtype)
+            self.model(probe_input)
         self.built = True
+        self.input_shape = list(input_shape)
         return self
 
     def _check_keras_layer(self):
-        if self.has_keras_layer and not self.built:
-            raise AssertionError("You should call `model.build()` first because the model contains `KerasLayer`.")
+        if self._has_keras_layer and not self.built:
+            raise AssertionError(
+                "You need to build the model via `.build()` before this operation. Because it contains `KerasLayer`."
+                )
 
-    def summary(self, input_shape, depth=3):
+    def summary(self, input_shape=None, depth=3, device=None):
         """Print a string summary of the network."""
         self._check_keras_layer()
-        torchsummary.summary(self.model, input_shape, depth=depth, verbose=1)
+        if input_shape is None:
+            input_shape = self.input_shape
+        assert (input_shape is not None)
+
+        if device is None:
+            device = self.trainer.device if self.compiled else 'cpu'
+
+        summary(self.model, input_shape, depth=depth, verbose=1, device=device)
 
     def compile(self, optimizer, loss, metrics=None, device=None):
         """
@@ -82,7 +99,7 @@ class Model(torch.nn.Module):
             device = 'cuda' if torch.cuda.is_available() else 'cpu'
             
         loss = _create_loss(loss)
-        optimizer = _create_optimizer(optimizer, self.parameters())
+        optimizer = _create_optimizer(optimizer, self.trainable_params())
 
         m = OrderedDict({'loss': loss})
         if isinstance(metrics, dict):
@@ -108,39 +125,64 @@ class Model(torch.nn.Module):
                 val_loader=None,
                 callbacks=[],
                 verbose=1,
-                precise_train_metrics=False):
+                use_amp=False):
 
+        assert self.compiled
         self.trainer.register_callbacks(callbacks)
-        history = self.trainer.run(train_loader, val_loader, max_epochs=epochs, verbose=verbose, precise_train_metrics=precise_train_metrics)
+        history = self.trainer.run(train_loader, val_loader, max_epochs=epochs, verbose=verbose, use_amp=use_amp)
 
         return history
 
 
     def fit(self, x, y, epochs, batch_size=32,
-                validation_split=None, val_split_seed=7,
+                validation_split=None, shuffle_val_split=True,
                 validation_data=None,
                 callbacks=[],
                 verbose=1,
-                precise_train_metrics=False,
                 shuffle=True,
                 sample_weight=None,
-                num_workers=0
+                num_workers=0,
+                use_amp=False
                 ):
-        """Train the model for a fixed number of epochs (iterations on a dataset)."""
+        """
+        Train the model for a fixed number of epochs (iterations on a dataset).
+
+        Args:
+
+        * `x` (`ndarray` or `torch.Tensor`): Input data 
+
+        * `y` (`ndarray` or `torch.Tensor`): Target data
+
+        * `epochs` (int): Number of epochs to train the model
+
+        * `batch_size` (int, default=32): Number of samples per gradient update
+
+        * `validation_split` (float between 0 and 1): Fraction of the training data to be used as validation data
+
+        * `shuffle_val_split` (bool, default=True): Whether to do shuffling when `validation_split` is provided
+
+        * `validation_data` (tuple of `x` and `y`): Data on which to evaluate the loss and any model metrics at the end of each epoch
+        
+        * `callbacks` (list of `keras4torch.callbacks.Callback`): List of callbacks to apply during training
+
+        * `verbose` (int, default=1): 0, 1, or 2. Verbosity mode. 0 = silent, 1 = normal, 2 = compat
+
+        * `shuffle` (bool, default=True): Whether to shuffle the training data before each epoch
+
+        * `sample_weight` (list of floats): Optional weights for the training samples. If provided will enable `WeightedRandomSampler`
+        
+        * `num_workers` (int, default=0): Workers of `DataLoader`. If `-1` will use `cpu_count() - 1` for multiprocessing
+
+        * `use_amp` (bool, default=False): Whether to use automatic mixed precision
+        """
 
         assert self.compiled
         x, y = to_tensor(x, y)
 
         assert not (validation_data != None and validation_split != None)
         has_val = validation_data != None or validation_split != None
-
-        if type(sample_weight) != type(None):
-            if isinstance(sample_weight, list):
-                sample_weight = torch.tensor(sample_weight)
-            sample_weight = to_tensor(sample_weight).float()
-            train_set = TensorDataset(x, y, sample_weight)
-        else:
-            train_set = TensorDataset(x, y)
+        
+        train_set = TensorDataset(x, y)
 
         if validation_data != None:
             x_val, y_val = to_tensor(validation_data[0], validation_data[1])
@@ -148,29 +190,38 @@ class Model(torch.nn.Module):
 
         if validation_split != None:
             val_length = int(len(train_set) * validation_split)
-            train_length = len(train_set) - val_length
-            train_set, val_set = random_split(train_set, [train_length, val_length], generator=torch.Generator().manual_seed(val_split_seed))
+            idx = np.arange(0, len(train_set))
+            if shuffle_val_split:
+                np.random.seed(1234567890)
+                np.random.shuffle(idx)
+            train_set, val_set = Subset(train_set, idx[:-val_length]), Subset(train_set, idx[-val_length:])
+            if sample_weight is not None:
+                assert len(sample_weight) == len(x)
+                sample_weight = [sample_weight[i] for i in idx[:-val_length]]
 
-        train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=shuffle, pin_memory=True, num_workers=num_workers)
-        val_loader = DataLoader(val_set, batch_size=batch_size, shuffle=False, pin_memory=True, num_workers=num_workers) if has_val else None
+        if sample_weight is not None:
+            sampler = WeightedRandomSampler(sample_weight, len(sample_weight))
+            shuffle = None
+        else:
+            sampler = None
+        
+        dl_kwargs = {'batch_size': batch_size, 'num_workers': _get_num_workers(num_workers)}
+        train_loader = DataLoader(train_set, shuffle=shuffle, sampler=sampler, **dl_kwargs)
+        val_loader = DataLoader(val_set, shuffle=False, **dl_kwargs) if has_val else None
 
-        # Training
-        self.trainer.register_callbacks(callbacks)
-        history = self.trainer.run(train_loader, val_loader, max_epochs=epochs, verbose=verbose, precise_train_metrics=precise_train_metrics)
-
-        return history
+        return self.fit_dl(train_loader, epochs, val_loader, callbacks, verbose, use_amp)
 
     @torch.no_grad()
-    def evaluate(self, x, y, batch_size=32, num_workers=0):
+    def evaluate(self, x, y, batch_size=32, num_workers=0, use_amp=False):
         """Return the loss value & metrics values for the model in test mode.\n\n    Computation is done in batches."""
         x, y = to_tensor(x, y)
-        val_loader = DataLoader(TensorDataset(x, y), batch_size=batch_size, shuffle=False, pin_memory=True, num_workers=num_workers)
-        return self.evaluate_dl(val_loader)
+        val_loader = DataLoader(TensorDataset(x, y), batch_size=batch_size, shuffle=False, num_workers=_get_num_workers(num_workers))
+        return self.evaluate_dl(val_loader, use_amp)
 
     @torch.no_grad()
-    def evaluate_dl(self, data_loader):
+    def evaluate_dl(self, data_loader, use_amp=False):
         assert self.compiled
-        return self.trainer.evaluate(data_loader)
+        return self.trainer.evaluate(data_loader, use_amp)
 
     @torch.no_grad()
     def predict_dl(self, data_loader, device=None, output_numpy=True, activation=None):
@@ -182,9 +233,10 @@ class Model(torch.nn.Module):
                 device = 'cuda' if torch.cuda.is_available() else 'cpu'
         self.eval().to(device=device)
 
-        outputs = [self.forward(x_batch[0].to(device=device)) for x_batch in data_loader]
+        outputs = [self(x_batch[0].to(device=device)) for x_batch in data_loader]
         outputs = torch.cat(outputs, dim=0)
 
+        activation = _create_activation(activation)
         if activation != None:
             outputs = activation(outputs)
 
@@ -196,7 +248,7 @@ class Model(torch.nn.Module):
         Generate output predictions for the input samples.\n\n    Computation is done in batches.
         """
         inputs = to_tensor(inputs)
-        data_loader = DataLoader(TensorDataset(inputs), batch_size=batch_size, shuffle=False, pin_memory=True, num_workers=num_workers)
+        data_loader = DataLoader(TensorDataset(inputs), batch_size=batch_size, shuffle=False, num_workers=_get_num_workers(num_workers))
         return self.predict_dl(data_loader, device=device, output_numpy=output_numpy, activation=activation)
 
     def save_weights(self, filepath):
