@@ -1,16 +1,14 @@
 import torch
-import pandas as pd
 import numpy as np
+from multiprocessing import cpu_count
 
 def to_tensor(*args):
     rt = []
     for arg in args:
-        if isinstance(arg, pd.DataFrame):
-            arg = torch.from_numpy(arg.values)
-        elif isinstance(arg, np.ndarray):
+        if isinstance(arg, np.ndarray):
             arg = torch.from_numpy(arg)
         elif not isinstance(arg, torch.Tensor):
-            raise TypeError('Only pd.DataFrame, np.ndarray and torch.Tensor are supported.')
+            raise TypeError('Only np.ndarray and torch.Tensor are supported.')
 
         if arg.dtype == torch.float64:
             print('[Warning] Auto convert float64 to float32, this could lead to extra memory usage.')
@@ -21,4 +19,11 @@ def to_tensor(*args):
 
         rt.append(arg)
                 
-    return rt[0] if len(rt) == 1 else tuple(rt)
\ No newline at end of file
+    return rt[0] if len(rt) == 1 else tuple(rt)
+
+
+def _get_num_workers(num_workers):
+    if num_workers == -1:
+        return cpu_count() - 1
+    else:
+        return min(num_workers, cpu_count())
\ No newline at end of file
